/////////////////////////////////////////////////////////////////////////////
// Name:        TASApp.cpp
// Purpose:     Defines the main application class
// Author:      Matthew J. Patitz
// Copyright:   Copyright 2008, 2009 Matthew J. Patitz
// Licence:
//    This file is part of ISU TAS.
//
//    ISU TAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    ISU TAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with ISU TAS.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////////////

// ----------------------------------------------------------------------------
// headers
// ----------------------------------------------------------------------------
#include "stdafx.h"
#include "wxprec.h"
#include <wx/datetime.h>
#include <wx/splash.h>
#include <wx/image.h>
#include <wx/xrc/xmlres.h>
#include <wx/file.h>
#include "TASSimFrm.h"
#include "TASApp.h"
#include "resource.h"
#include "TAS.h"
#include "TASSimFrm.h"
#include "TASEditor.h"
#include "TASDockers.h"
#include "TASTiles.h"
#include "TASTilePanel.h"
#include "TASDebug.h"
#include "RandomLib/Random.hpp"


TASApp*      g_pTASApp = NULL;
TASContext*  g_pContext;

extern void TASInitXmlResource(); // defined in generated file

TASContext::TASContext()
              : wxConfig(_T("TAS")),
                m_setSimWin(TASSimWinName, wxPoint(20, 20), wxSize(700, 700), true, false),
                m_setEditorWin(TASEditorWinName, wxPoint(320, 20), wxSize(700, 700), true, false)
{
  m_iTASDimMode           = TAS_2D_MODE;      //  Default to 2D mode
  m_iTASOptMode           = TAS_OPT_MODE_1;   //  Default to optimized
  m_iTASOptModeThreshold  = 500;              //  Try 500 for now...
  m_maskSimTileDisplay    = TAS_MASK_DISPALL;
  m_maskEditorTileDisplay = TAS_MASK_DISPALL;
  m_bUseEditor            = true;
  m_bOutputDebugMsgs      = false;
  m_bReportNondeterminism = false;
  m_bReport1stCondLD      = false;
  m_bMaxTileAdditions     = true;
  m_bAsyncDisplay         = false;
  m_iTASTileConcMode      = TAS_TILECONC_MODE_OFF;
}

void TASContext::LoadSavedSettings()
{
  //  Get saved settings for dimension mode
#ifdef TAS_BUILD_3D
  m_iTASDimMode = Read(TASDimModeName, (long)TAS_2D_MODE);
#else
  m_iTASDimMode = TAS_2D_MODE;
#endif  // TAS_BUILD_3D

  m_fnTilesetFile       = Read(_T("tdsFile"),  DefaultTilesetFileName);
  m_fnTilespaceFile     = Read(_T("tdpFile"),  DefaultTilespaceFileName);

  m_colSpaceBkgrnd.Set(Read(_T("bgColor"),   _T("#F3F77D")));//"#000000"));
  m_colDefaultTile.Set(Read(_T("tileColor"), _T("#FFFFFF")));
  m_colDefaultText.Set(Read(_T("textColor"), _T("#000000")));

  m_maskSimTileDisplay    = Read(_T("simTileDisplayMask"),    TAS_MASK_DISPALL);
  m_maskEditorTileDisplay = Read(_T("editorTileDisplayMask"), TAS_MASK_DISPALL);

/*
  m_sCellValuePaneInfo    = Read(_T("cellValPanePersp"),  _T(""));
  m_sSpaceConfigPaneInfo  = Read(_T("spacePanePersp"),    _T(""));
  m_sAxesPaneInfo         = Read(_T("axesPanePersp"),     _T(""));
  m_sOverviewPaneInfo     = Read("overviewPanePersp",     _T(""));
*/

  m_sSimFramePersp          = Read(_T("simFramePersp"),     _T(""));
  m_sEditorFramePersp       = Read(_T("editorFramePersp"),    _T(""));

  m_setSimWin.LoadSettings(this);
  m_setEditorWin.LoadSettings(this);

  Read(TASUseEditor, &m_bUseEditor, true);
  Read(TASOutputDebugMsgs, &m_bOutputDebugMsgs, false);
  Read(TASReportNondeterm, &m_bReportNondeterminism, false);
  Read(TASReport1stCondLD, &m_bReport1stCondLD, false);
  Read(TASMaxTileAdditions,&m_bMaxTileAdditions, false);
//  Read(TASMaxTileAdditions,&m_bAsyncDisplay, false);

  m_bDoSave = true;
}

void TASContext::SaveCurrentSettings()
{
  if (m_fnTilesetFile.IsRelative()) {
    m_fnTilesetFile.MakeAbsolute();
  }
  if (m_fnTilespaceFile.IsRelative()) {
    m_fnTilespaceFile.MakeAbsolute();
  }
  Write(_T("tdsFile"),  m_fnTilesetFile.GetFullPath());
  Write(_T("tdpFile"),  m_fnTilespaceFile.GetFullPath());
  Write(TASDimModeName, m_iTASDimMode);

  Write(_T("bgColor"), m_colSpaceBkgrnd.GetAsString(wxC2S_HTML_SYNTAX));
  Write(_T("tileColor"), m_colDefaultTile.GetAsString(wxC2S_HTML_SYNTAX));
  Write(_T("textColor"), m_colDefaultText.GetAsString(wxC2S_HTML_SYNTAX));

  Write(_T("simTileDisplayMask"),     (long)m_maskSimTileDisplay);
  Write(_T("editorTileDisplayMask"),  (long)m_maskEditorTileDisplay);

  m_setSimWin.SaveCurrentSettings(this);
  m_setEditorWin.SaveCurrentSettings(this);

/*
  //  Set best sizes to be the current sizes, then get pane info for each dockable window
  wxRect r1 = g_pTASApp->m_pSimFrame->m_mgr.GetPane(g_pTASApp->m_pSimFrame->m_pCellValueWin).rect;
  g_pTASApp->m_pSimFrame->m_mgr.GetPane(g_pTASApp->m_pSimFrame->m_pCellValueWin).BestSize(r1.width, r1.height);
  m_sCellValuePaneInfo = g_pTASApp->m_pSimFrame->m_mgr.SavePaneInfo(g_pTASApp->m_pSimFrame->m_mgr.GetPane(g_pTASApp->m_pSimFrame->m_pCellValueWin));

  wxRect r3 = g_pTASApp->m_pSimFrame->m_mgr.GetPane(g_pTASApp->m_pSimFrame->m_pSpaceConfigWin).rect;
  g_pTASApp->m_pSimFrame->m_mgr.GetPane(g_pTASApp->m_pSimFrame->m_pSpaceConfigWin).BestSize(r3.width, r3.height);
  m_sSpaceConfigPaneInfo = g_pTASApp->m_pSimFrame->m_mgr.SavePaneInfo(g_pTASApp->m_pSimFrame->m_mgr.GetPane(g_pTASApp->m_pSimFrame->m_pSpaceConfigWin));

  wxRect r4 = g_pTASApp->m_pSimFrame->m_mgr.GetPane(g_pTASApp->m_pSimFrame->m_pAxesWin).rect;
  g_pTASApp->m_pSimFrame->m_mgr.GetPane(g_pTASApp->m_pSimFrame->m_pAxesWin).BestSize(r4.width, r4.height);
  m_sAxesPaneInfo = g_pTASApp->m_pSimFrame->m_mgr.SavePaneInfo(g_pTASApp->m_pSimFrame->m_mgr.GetPane(g_pTASApp->m_pSimFrame->m_pAxesWin));

  wxRect r5 = g_pTASApp->m_pSimFrame->m_mgr.GetPane(g_pTASApp->m_pSimFrame->m_pOverviewWin).rect;
  g_pTASApp->m_pSimFrame->m_mgr.GetPane(g_pTASApp->m_pSimFrame->m_pOverviewWin).BestSize(r5.width, r5.height);
  m_sOverviewPaneInfo = g_pTASApp->m_pSimFrame->m_mgr.SavePaneInfo(g_pTASApp->m_pSimFrame->m_mgr.GetPane(g_pTASApp->m_pSimFrame->m_pOverviewWin));

  Write(_T("cellValPanePersp"), m_sCellValuePaneInfo);
  Write(_T("spacePanePersp"),   m_sSpaceConfigPaneInfo);
  Write(_T("axesPanePersp"),    m_sAxesPaneInfo);
  Write(_T("overviewPanePersp"),m_sOverviewPaneInfo);
*/

  m_sSimFramePersp        = g_pTASApp->m_pSimFrame->m_mgr.SavePerspective();
  Write(_T("simFramePersp"),   m_sSimFramePersp);

  if (g_pTASApp->m_pEditorFrame) {
    m_sEditorFramePersp       = g_pTASApp->m_pEditorFrame->m_mgr.SavePerspective();
    Write(_T("editorFramePersp"),   m_sEditorFramePersp);
  }

  Write(TASUseEditor, m_bUseEditor);
  Write(TASOutputDebugMsgs, m_bOutputDebugMsgs);
  Write(TASReportNondeterm, m_bReportNondeterminism);
  Write(TASReport1stCondLD, m_bReport1stCondLD);
  Write(TASMaxTileAdditions,m_bMaxTileAdditions);
//  Write(TASMaxTileAdditions,m_bAsyncDisplay);
}

bool TASContext::IsAssemblyFileOK()
{
  if ((m_fnTilespaceFile.GetFullName().Len() > 0) &&
      (m_fnTilespaceFile.GetFullName() != DefaultTilespaceFileName) &&
      (m_fnTilespaceFile.IsOk()))
  {
    return true;
  }
  return false;
}

bool TASContext::IsTilesetFileOK()
{
  if ((m_fnTilesetFile.GetFullName().Len() > 0) &&
      (m_fnTilesetFile.GetFullName() != DefaultTilesetFileName) &&
      (m_fnTilesetFile.IsOk()))
  {
    return true;
  }
  return false;
}

TASWinSettings::TASWinSettings(wxString strName)
{
  m_pWin        = NULL;
  m_strName     = strName;
  m_ptLoc       = wxDefaultPosition;
  m_szSize      = wxDefaultSize;
  m_bVisible    = true;
  m_bMaximized  = false;
}

TASWinSettings::TASWinSettings(wxString strName, wxPoint ptLoc, wxSize szSize, bool bVisible, bool bMaximized)
{
  m_pWin        = NULL;
  m_strName     = strName;
  m_ptLoc       = ptLoc;
  m_szSize      = szSize;
  m_bVisible    = bVisible;
  m_bMaximized  = bMaximized;
}

TASWinSettings& TASWinSettings::operator=(const TASWinSettings& rhs)
{
  m_strName     = rhs.m_strName;
  m_ptLoc       = rhs.m_ptLoc;
  m_szSize      = rhs.m_szSize;
  m_bVisible    = rhs.m_bVisible;
  m_bMaximized  = rhs.m_bMaximized;

  return *this;
}

void TASWinSettings::GetCurrentSettings()
{
  if (m_pWin) {
    if (m_pWin->IsShown()) {
      m_bVisible = true;
    } else {
      m_bVisible = false;
    }

    m_bMaximized = m_pWin->IsMaximized();
    if (m_bMaximized) {
      m_pWin->Maximize(false);
    }

    m_ptLoc.x = m_pWin->GetPosition().x;
    m_ptLoc.y = m_pWin->GetPosition().y;
    m_szSize.SetWidth(m_pWin->GetSize().GetWidth());
    m_szSize.SetHeight(m_pWin->GetSize().GetHeight());

  } else {
    m_bVisible = false;
  }
}

void TASWinSettings::SaveCurrentSettings(wxConfig* pConfig)
{
  GetCurrentSettings();

  pConfig->Write(m_strName + _T("LocX"), m_ptLoc.x);
  pConfig->Write(m_strName + _T("LocY"), m_ptLoc.y);
  pConfig->Write(m_strName + _T("SizeX"), m_szSize.GetWidth());
  pConfig->Write(m_strName + _T("SizeY"), m_szSize.GetHeight());
  pConfig->Write(m_strName + _T("Vis"), (m_bVisible ? 1 : 0));
  pConfig->Write(m_strName + _T("Maxd"), (m_bMaximized ? 1 : 0));
}

void TASWinSettings::LoadSettings(wxConfig* pConfig)
{
  m_ptLoc.x = pConfig->Read(m_strName + _T("LocX"), m_ptLoc.x);
  m_ptLoc.y = pConfig->Read(m_strName + _T("LocY"), m_ptLoc.y);
  m_szSize.SetWidth(pConfig->Read(m_strName + _T("SizeX"), m_szSize.GetWidth()));
  m_szSize.SetHeight(pConfig->Read(m_strName + _T("SizeY"), m_szSize.GetHeight()));
  m_bVisible = ((pConfig->Read(m_strName + _T("Vis"), (long)1) == 0) ? false : true);
  m_bMaximized = ((pConfig->Read(m_strName + _T("Maxd"), (long)0) == 0) ? false : true);
}


// Create a new application object: this macro will allow wxWindows to create
// the application object during program execution (it's better than using a
// static object for many reasons) and also declares the accessor function
// wxGetApp() which will return the reference of the right type (i.e. TASApp and
// not wxApp)
IMPLEMENT_APP(TASApp)

// ============================================================================
// implementation
// ============================================================================

// ----------------------------------------------------------------------------
// the application class
// ----------------------------------------------------------------------------

static wxCriticalSection* s_critSect_logMsg = NULL;

wxMutex* g_pMutexTileSpaceGuard = NULL;

TASApp::TASApp()
{
  //  Create the mutex that will be used to protect against concurrent accesses
  //  to the tile space by multiple threads
  g_pMutexTileSpaceGuard   = new wxMutex();

  m_pSimFrame       = NULL;
  m_pEditorFrame    = NULL;
  m_bShowSplash     = false;        // Don't show splash screen for now since it is very obsolete

  Srand(0, true);
  m_randVal         = m_randSeed;

  g_pTASApp         = this;
}

TASApp::~TASApp()
{
  if (g_pMutexTileSpaceGuard) {
    delete g_pMutexTileSpaceGuard;
  }

  if (s_critSect_logMsg) {
    delete s_critSect_logMsg;
  }
}

void TASApp::SetTASDimMode(int mode)
{
  g_pContext->m_iTASDimMode = mode;
}

int  TASApp::GetTASDimMode()
{
  return g_pContext->m_iTASDimMode;
}

// 'Main program' equivalent: the program execution "starts" here
bool TASApp::OnInit()
{
  // Resource initialization
  wxImage::AddHandler(new wxGIFHandler);
  wxImage::AddHandler(new wxICOHandler);
  wxImage::AddHandler(new wxXPMHandler);
  wxXmlResource::Get()->InitAllHandlers();
  TASInitXmlResource();

  g_pContext = new TASContext();

  //  Create the global tileset and tile space objects
  NewTAS();

  g_pContext->LoadSavedSettings();

  ProcessCommandLineFiles(argc, argv);

  // create the main TAS window
  m_pSimFrame = new TASSimulatorFrame(ID_TAS_SIMULATOR_FRAME, this, _("TAS 2.0"), g_pContext->m_setSimWin.m_ptLoc, 
                                      g_pContext->m_setSimWin.m_szSize, wxDEFAULT_FRAME_STYLE);
  m_pSimFrame->Maximize(g_pContext->m_setSimWin.m_bMaximized);
  g_pContext->m_setSimWin.m_pWin = m_pSimFrame;

  if (g_pContext->m_bUseEditor) {
    CreateEditor();
  }

  m_pSimFrame->Show(TRUE);
  SetTopWindow(m_pSimFrame);

  m_pSimFrame->Refresh();
  m_pSimFrame->Update();

  if (m_bShowSplash) {
    // create splash screen
    wxBitmap bitmap;// = wxXmlResource::Get()->LoadBitmap("program.gif");
    if (bitmap.Ok())
    {
      wxSplashScreen* splash = new wxSplashScreen(bitmap, wxSPLASH_CENTRE_ON_SCREEN|wxSPLASH_TIMEOUT,
                                                  6000, NULL, -1, wxDefaultPosition, wxDefaultSize,
                                                  wxSTAY_ON_TOP);
    }
  }

///////MJP: stuff to wrap up into a separate function
  if (g_pContext->IsTilesetFileOK()) {
    if (ReloadTilesetIntoEditor()) {
      if (LoadTilesetIntoSimulator() &&
          g_pContext->IsAssemblyFileOK())
      {
        ReloadAssembly();
      }
    }
  }

  m_pSimFrame->m_panel->SetViewableSpace(TASCoords(0, 0, 0), TASCoords(0, 0, 0), TAS_3D_VIEWBY_UNLIMITED);
  if (m_pSimFrame->m_pCellValueWin) {
    m_pSimFrame->m_pCellValueWin->SetFields();
  }
  if (m_pSimFrame->m_pSpaceConfigWin) {
    m_pSimFrame->m_pSpaceConfigWin->SetViewType(TAS_3D_VIEWBY_UNLIMITED);
    m_pSimFrame->m_pSpaceConfigWin->SetCurrentDims();
  }

///////////

  if (GetTASDimMode() == TAS_2D_MODE) {
    ((TASTilePanel*)m_pSimFrame->m_panel)->CheckCanvasSize();
    m_pSimFrame->m_panel->ScrollToCenter(g_pTileSpace->m_seed.m_coordsLeftmost);
    m_pSimFrame->m_panel->UpdateSpace(true);
//    m_pEditorFrame->ProcessUpdatedTileset();
  }

//MJP: an unworking hack to try to fix the fact that the 3-D canvas doesn't always fill the panel
//m_pSimFrame->SetTASDimMode(TAS_3D_MODE, true);

  // success: wxApp::OnRun() will be called which will enter the main message
  // loop and the application will run. If we returned FALSE here, the
  // application would exit immediately.
  return TRUE;
}

bool TASApp::CreateEditor()
{
  // create the editor window
  m_pEditorFrame = new TASEditorFrame(ID_TAS_EDITOR_FRAME, this, g_pContext->m_setEditorWin.m_ptLoc, g_pContext->m_setEditorWin.m_szSize);
  m_pEditorFrame->Maximize(g_pContext->m_setEditorWin.m_bMaximized);
  g_pContext->m_setEditorWin.m_pWin = m_pEditorFrame;

  m_pEditorFrame->Show(TRUE);

  return true;
}

size_t GetTileTypeCount(const wxString& str)
{
  size_t count  = 0;
  size_t pos    = 0;
  while (pos < str.length()) {
    size_t loc = str.find(_T("TILENAME "), pos);
    if (loc != std::string::npos) {
      count++;
      pos = loc + wxStrlen(_T("TILENAME "));
    } else {
      break;
    }
  }
  return count;
}

wxString GetTdsFileNameFromTdpFile(wxString& sTASFileName)
{
  wxString sRet;

  wxFFile f(sTASFileName);
  if (f.IsOpened()) {
    wxChar sBuff[501] = {0};
    if (f.Read(sBuff, 500)) {
      wxString sContents(sBuff);
      if (sContents.Find(wxT(".tds")) != wxNOT_FOUND) {
        sRet = sContents.SubString(0, sContents.Find(wxT(".tds"))-1) + wxT(".tds");
      }
    }
    f.Close();
  }
  return sRet;
}

class TASLoaderThread : public wxThread
{
public:
  TASLoaderThread(wxString sTdsFileName, wxString sTdpFileName, bool bLoadIntoSim);

  void* Entry();

  wxString m_sTdsFileName;
  wxString m_sTdpFileName;
  bool     m_bLoadIntoSim;
  bool     m_bReturnVal;
};

TASLoaderThread::TASLoaderThread(wxString sTdsFileName, wxString sTdpFileName, bool bLoadIntoSim)
                : wxThread(wxTHREAD_JOINABLE)
{
  m_sTdsFileName = sTdsFileName;
  m_sTdpFileName = sTdpFileName;
  m_bLoadIntoSim = bLoadIntoSim;

  m_bReturnVal   = false;
}

void* TASLoaderThread::Entry()
{
  m_bReturnVal = true;
  if (!m_sTdsFileName.IsEmpty()) {
    m_bReturnVal = g_pTASApp->LoadTilesetIntoEditor(m_sTdsFileName);
    if (m_bReturnVal && m_bLoadIntoSim) {
      m_bReturnVal = g_pTASApp->LoadTilesetIntoSimulator();
    }
  }
  if (m_bReturnVal && !m_sTdpFileName.IsEmpty()) {
    m_bReturnVal = g_pTASApp->LoadAssembly(m_sTdpFileName);
  }
  return (void*)&m_bReturnVal;
}


bool TASApp::LoadTAS(wxString& sTASFileName)
{
  wxString sTdsFileName = GetTdsFileNameFromTdpFile(sTASFileName);
  if (sTdsFileName == _T("")) {
    sTdsFileName = sTASFileName;
    sTdsFileName.Replace(_T(".tdp"), _T(".tds"));
  }
/*
  TASLoaderThread threadLoader(sTdsFileName, sTASFileName, true);
  threadLoader.Create();
  threadLoader.Run();
  threadLoader.Wait();

  return true;
*/
  wxFileName fnTds(sTdsFileName);
  wxFileName fnTdp(sTASFileName);
  if (fnTds.IsRelative()) {
    fnTds.Assign(fnTdp.GetPath(), fnTds.GetName(), fnTds.GetExt());
  }
  sTdsFileName = fnTds.GetFullPath();
  wxString sTdpFileName = fnTdp.GetFullPath();
  return (LoadTilesetIntoEditor(sTdsFileName) && LoadTilesetIntoSimulator() && LoadAssembly(sTdpFileName));
}

void TASApp::SetDimModeFromTileset()
{
  int mode = TAS_2D_MODE;
  for (size_t i=0; i<g_pEditorTileset->size(); i++) {
    if ((g_pEditorTileset->at(i)->m_sides[TAS_UP].m_iStrength != 0) ||
        (g_pEditorTileset->at(i)->m_sides[TAS_DOWN].m_iStrength != 0))
    {
      mode = TAS_3D_MODE;
      break;
    }
  }
  if (g_pContext->m_iTASDimMode != mode) {
    if (m_pSimFrame) {
      m_pSimFrame->SetTASDimMode(mode, true);
    }
    if (m_pEditorFrame) {
      m_pEditorFrame->SetDimMode();
      g_pEditorTileset->ReIndex();
      g_pTileSpace->m_tileset.ReIndex();
    }
  }
}

bool TASApp::LoadTilesetIntoEditor(wxString& sTilesetFileName)
{
  wxFileName fnTilesetFile(sTilesetFileName);
  wxFFile f(fnTilesetFile.GetFullName(), _T("r"));
  if (f.IsOpened()) {
    wxString sContents;
    if (f.ReadAll(&sContents)) {
      size_t ttCount = GetTileTypeCount(sContents);

      wxString sMsg = wxString::Format(_T("%u tiletypes being loaded into editor...\n"), ttCount); 
      SetEditorOutputText(sMsg, false, false);

      if (ttCount > g_pContext->m_iTASOptModeThreshold) {
        g_pContext->m_iTASOptMode = TAS_OPT_MODE_0;

          wxString sMsg2 = wxString::Format(_T("This tileset is very large, so several editor enhancements are disabled.\n")); 
          SetEditorOutputText(sMsg2, true, false);

      } else {
        g_pContext->m_iTASOptMode = TAS_OPT_MODE_1;
      }
      *g_pEditorTileset = sContents;
      f.Close();
      g_pContext->m_fnTilesetFile = sTilesetFileName;
      SetDimModeFromTileset();
      if (m_pEditorFrame) {
        m_pEditorFrame->ProcessUpdatedTileset();
      }
      return true;
    }
    f.Close();
  }
  return false;
}

bool TASApp::LoadTilesetFromSimulator()
{
  wxString sMsg = wxString::Format(_T("%u tiletypes being loaded into editor...\n"), g_pTileSpace->m_tileset.size()); 
  SetEditorOutputText(sMsg, false, false);

  if (g_pTileSpace->m_tileset.size() > g_pContext->m_iTASOptModeThreshold) {
    g_pContext->m_iTASOptMode = TAS_OPT_MODE_0;

    wxString sMsg2 = wxString::Format(_T("This tileset is very large, so several editor enhancements are disabled.\n")); 
    SetEditorOutputText(sMsg2, true, false);

  } else {
    g_pContext->m_iTASOptMode = TAS_OPT_MODE_1;
  }
  ((TASTileTypeSet&)*g_pEditorTileset) = g_pTileSpace->m_tileset;
  if (m_pEditorFrame) {
    m_pEditorFrame->ProcessUpdatedTileset();
  }
  g_pEditorTileset->SetSimMod(false);
  return true;
}

bool TASApp::LoadTilesetIntoSimulator()
{
  wxString sMsg = wxString::Format(_T("%u tiletypes being loaded into simulator...\n"), g_pEditorTileset->size()); 
  SetSimOutputText(sMsg, false, false);

  if (g_pContext->m_iTASOptMode == TAS_OPT_MODE_0) {
    wxString sMsg2 = wxString::Format(_T("This tileset is very large, so some seed assembly editing enhancements are disabled.\n")); 
    SetSimOutputText(sMsg2, true, false);
  }
  g_pTileSpace->m_tileset = ((TASTileTypeSet&)*g_pEditorTileset);
  g_pEditorTileset->SetSimMod(false);
  m_pSimFrame->ProcessUpdatedTileset();
  return true;
}

bool TASApp::LoadAssembly(wxString& sAssemblyFileName)
{
  wxFileName fnAssemblyFile(sAssemblyFileName);
  wxFFile f(fnAssemblyFile.GetFullName(), _T("r"));
  if (f.IsOpened()) {
    bool bRet = g_pTileSpace->LoadSpaceDefinitionFromFile(&f);
    if (bRet) {
      g_pContext->m_fnTilespaceFile = sAssemblyFileName;
      if (m_pSimFrame) {
        m_pSimFrame->ProcessUpdatedTilespace();
      }
    }
    f.Close();
    return bRet;
  }
  return false;
}

bool TASApp::ReloadTAS()
{
  return (ReloadTilesetIntoEditor() && LoadTilesetIntoSimulator() && ReloadAssembly());
}

bool TASApp::ReloadTilesetIntoEditor()
{
  if (g_pContext->IsTilesetFileOK()) {
    wxFFile f;
    if (f.Open(g_pContext->m_fnTilesetFile.GetFullPath())) {
      wxString sContents;
      if (f.ReadAll(&sContents)) {
        size_t ttCount = GetTileTypeCount(sContents);

        wxString sMsg = wxString::Format(_T("%u tiletypes being loaded into editor...\n"), ttCount);
        SetEditorOutputText(sMsg, false, false);

        if (ttCount > g_pContext->m_iTASOptModeThreshold) {
          g_pContext->m_iTASOptMode = TAS_OPT_MODE_0;

          wxString sMsg2 = wxString::Format(_T("This tileset is very large, so several editor enhancements are disabled.\n")); 
          SetEditorOutputText(sMsg2, true, false);

        } else {
          g_pContext->m_iTASOptMode = TAS_OPT_MODE_1;
        }
        *g_pEditorTileset = sContents;
        f.Close();
        SetDimModeFromTileset();
        if (m_pEditorFrame) {
          m_pEditorFrame->ProcessUpdatedTileset();
        }
        return true;
      }
      f.Close();
    }
  }
  return false;
}

bool TASApp::ReloadAssembly()
{
  if (g_pContext->IsAssemblyFileOK()) {
    wxFFile f;
    if (f.Open(g_pContext->m_fnTilespaceFile.GetFullPath())) {
      bool bRet = g_pTileSpace->LoadSpaceDefinitionFromFile(&f);
      if (bRet) {
        if (m_pSimFrame) {
          m_pSimFrame->ProcessUpdatedTilespace();
        }
      }
      f.Close();
      return bRet;
    }
  }
  return false;
}

bool TASApp::ResetTAS()
{
  return (/*LoadTilesetIntoEditor() &&*/ResetAssembly());
}

bool TASApp::ResetAssembly()
{
  g_pTileSpace->ResetSpace(true);
  if (m_pSimFrame) {
    m_pSimFrame->ProcessUpdatedTilespace();
  }
  return true;
}

bool TASApp::SaveTASAs(wxString& sTASFileName)
{
  if (SaveAssemblyAs(sTASFileName)) {
    wxString sTds = sTASFileName;
    sTds.Replace(_T(".tdp"), _T(".tds"));
    return SaveTilesetAs(sTds);
  }
  return false;
}

bool TASApp::SaveTilesetAs(wxString& sTilesetFileName)
{
  wxFFile f(sTilesetFileName, _T("wb"));
  if (f.IsOpened()) {
    g_pContext->m_fnTilesetFile = sTilesetFileName;
    bool bRet = g_pEditorTileset->SaveToFile(&f);
    f.Close();
    return bRet;
  } else {
    wxString errMsg = wxString::Format(_T("There has been an error saving the tileset to the file:\n'%s'.\nPlease check to make sure the file permissions allow this."), sTilesetFileName.c_str());
    wxMessageBox(errMsg, _T("Error saving tileset"), wxICON_ERROR);
  }
  return false;
}

bool TASApp::SaveAssemblyAs(wxString& sAssmblyFileName)
{
  wxFFile f(sAssmblyFileName, _T("wb"));
  if (f.IsOpened()) {
    g_pContext->m_fnTilespaceFile = sAssmblyFileName;
    m_pSimFrame->SetTitle();
    bool bRet = g_pTileSpace->SaveToFile(&f);
    f.Close();
    return bRet;
  } else {
    wxString errMsg = wxString::Format(_T("There has been an error saving the tile assembly to the file:\n'%s'.\nPlease check to make sure the file permissions allow this."), sAssmblyFileName.c_str());
    wxMessageBox(errMsg, _T("Error saving tile assembly"), wxICON_ERROR);
  }
  return false;
}

bool TASApp::SaveTAS()
{
  return (SaveTileset() && SaveAssembly());
}

bool TASApp::SaveTileset()
{
  if (g_pContext->IsTilesetFileOK()) {
    wxFFile f(g_pContext->m_fnTilesetFile.GetFullPath(), _T("wb"));
    if (f.IsOpened()) {
      bool bRet = g_pEditorTileset->SaveToFile(&f);
      f.Close();
      return bRet;
    } else {
      wxString errMsg = wxString::Format(_T("There has been an error saving the tile set to the file:\n'%s'.\nPlease check to make sure the file permissions allow this."), g_pContext->m_fnTilesetFile.GetFullPath().c_str());
      wxMessageBox(errMsg, _T("Error saving tile set"), wxICON_ERROR);
      return false;
    }
  } else {
    return false;
  }
}

bool TASApp::SaveAssembly()
{
  if (g_pContext->IsAssemblyFileOK()) {
    wxFFile f(g_pContext->m_fnTilespaceFile.GetFullPath(), _T("wb"));
    if (f.IsOpened()) {
      bool bRet = g_pTileSpace->SaveToFile(&f);
      f.Close();
      return bRet;
    } else {
      wxString errMsg = wxString::Format(_T("There has been an error saving the tile assembly to the file:\n'%s'.\nPlease check to make sure the file permissions allow this."), g_pContext->m_fnTilespaceFile.GetFullPath().c_str());
      wxMessageBox(errMsg, _T("Error saving tile assembly"), wxICON_ERROR);
      return false;
    }
  } else {
    return false;
  }
}

bool TASApp::NewTAS()
{
  return (NewTileset() && NewAssembly());
}

bool TASApp::NewTileset()
{
  g_pContext->m_fnTilesetFile = DefaultTilesetFileName;
  if (g_pEditorTileset) {
    delete g_pEditorTileset;
  }
  g_pContext->m_iTASOptMode = TAS_OPT_MODE_1;
  g_pEditorTileset = new TASTileTypeSetEx();
  if (m_pEditorFrame) {
    m_pEditorFrame->ProcessUpdatedTileset();
  }
  return true;
}

bool TASApp::NewAssembly()
{
  g_pContext->m_fnTilespaceFile = DefaultTilespaceFileName;
  if (g_pTileSpace) {
    delete g_pTileSpace;
  }
  g_pTileSpace = new TASTileSpace(g_pContext->m_iTASDimMode == TAS_2D_MODE ? 2 : 3);
  if (m_pSimFrame) {
    m_pSimFrame->ProcessUpdatedTilespace();
  }
  return true;
}

void TASApp::SetTemp(int temp)
{
  g_pTileSpace->SetTemp(temp);
  if (g_pTASApp && g_pTASApp->m_pEditorFrame && g_pTASApp->m_pEditorFrame->m_pTTWin) {
    g_pTASApp->m_pEditorFrame->m_pTTWin->SetTemps();
  }
  if (m_pSimFrame && m_pSimFrame->m_panel) {
    m_pSimFrame->m_panel->UpdateSpace(true);
  }
}

void TASApp::SetSimOutputText(wxString& sMsg, bool bAppend, bool bRaiseWin)
{
  if (m_pSimFrame) {
    wxCommandEvent evt(tasEVT_POST_SETOUTPUTTEXT, ID_TAS_SIMULATOR_FRAME);
    evt.SetString(sMsg);

    int flags = 0;
    if (bAppend)    flags = flags | TAS_APPEND_FLAG;
    if (bRaiseWin)  flags = flags | TAS_RAISEWIN_FLAG;
    evt.SetInt(flags);

    ::wxPostEvent(g_pTASApp->m_pSimFrame, evt);
  }
}

void TASApp::SetSimDebugText(wxString& sMsg, bool bAppend, bool bRaiseWin)
{
  if (m_pSimFrame) {
    wxCommandEvent evt(tasEVT_POST_SETDEBUGTEXT, ID_TAS_SIMULATOR_FRAME);
    evt.SetString(sMsg);

    int flags = 0;
    if (bAppend)    flags = flags | TAS_APPEND_FLAG;
    if (bRaiseWin)  flags = flags | TAS_RAISEWIN_FLAG;
    evt.SetInt(flags);

    ::wxPostEvent(g_pTASApp->m_pSimFrame, evt);
  }
}

void TASApp::SetEditorOutputText(wxString& sMsg, bool bAppend, bool bRaiseWin)
{
  if (m_pEditorFrame) {
    wxCommandEvent evt(tasEVT_POST_SETOUTPUTTEXT, ID_TAS_EDITOR_FRAME);
    evt.SetString(sMsg);

    int flags = 0;
    if (bAppend)    flags = flags | TAS_APPEND_FLAG;
    if (bRaiseWin)  flags = flags | TAS_RAISEWIN_FLAG;
    evt.SetInt(flags);

    ::wxPostEvent(g_pTASApp->m_pEditorFrame, evt);
  }
}

void TASApp::SetEditorDebugText(wxString& sMsg, bool bAppend, bool bRaiseWin)
{
  if (m_pEditorFrame) {
    wxCommandEvent evt(tasEVT_POST_SETDEBUGTEXT, ID_TAS_EDITOR_FRAME);
    evt.SetString(sMsg);

    int flags = 0;
    if (bAppend)    flags = flags | TAS_APPEND_FLAG;
    if (bRaiseWin)  flags = flags | TAS_RAISEWIN_FLAG;
    evt.SetInt(flags);

    ::wxPostEvent(g_pTASApp->m_pEditorFrame, evt);
  }
}


void displayHelp() {
  //MJP: TODO - to stdout or a dialog?
}

void TASApp::ProcessCommandLineFiles(int argc, wxChar* argv[])
{
}

void TASApp::SaveUserSettings()
{
  if (g_pContext->m_bDoSave) {
    g_pContext->SaveCurrentSettings();
  }
}

unsigned long TASApp::m_randSeed = 1L;
unsigned long TASApp::m_randVal  = 1L;

void TASApp::Srand(unsigned long seed, bool bRandomSeed)
{
  if (bRandomSeed) {
    RandomLib::Random::Global.Reseed();
  } else {
    RandomLib::Random::Global.Reseed(seed);
  }
  std::vector<RandomLib::RandomSeed::seed_type> vect = RandomLib::Random::Global.Seed();
  m_randSeed = vect[0];
}

/*
int TASApp::Rand()
{
//currently unimplemented
}
*/

int TASApp::RandInRange(int n) {
  return RandomLib::Random::Global.Integer(n);
}

int TASApp::OnExit()
{
  delete g_pTileSpace;
  delete g_pEditorTileset;
  delete g_pContext;

/*
  if (g_printData) {
    delete g_printData;
  }
  if (g_pageSetupData) {
    delete g_pageSetupData;
  }
*/

  return 0;
}

