/////////////////////////////////////////////////////////////////////////////
// Name:        TASTileCanvas.cpp
// Purpose:     Defines the panel class used for 2-D tile space rendering
// Author:      Matthew J. Patitz
// Copyright:   Copyright 2008, 2009 Matthew J. Patitz
// Licence:
//    This file is part of ISU TAS.
//
//    ISU TAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    ISU TAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with ISU TAS.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#ifdef __GNUG__
#pragma implementation
#pragma interface
#endif

// For compilers that support precompilation, includes "wx.h".
#include "wx/wxprec.h"

#ifdef __BORLANDC__
#pragma hdrstop
#endif

#ifndef WX_PRECOMP
#include "wx/wx.h"
#endif

#include <math.h>
#include <wx/xrc/xmlres.h>
#include "TASTileCanvas.h"
#include "TASTiles.h"
#include "TASDockers.h"
#include "TASApp.h"
#include <wx/dcbuffer.h>
#include "TASDebug.h"


#define TAS_DRAW_INVALID     -1
#define TAS_DRAW_EMPTY        0
#define TAS_DRAW_FRONTIER     1
#define TAS_DRAW_FRONTIERDEAD 2
#define TAS_DRAW_TILE         3

#define TAS_ZOOM_IN   1
#define TAS_ZOOM_OUT  2

#ifndef max
#define max(a, b) (((a) > (b)) ? (a) : (b))
#endif
#define RADS_IN_CIRCLE 6.28318531f


IMPLEMENT_CLASS(TASRightClickWin, wxPanel)
BEGIN_EVENT_TABLE(TASRightClickWin, wxPanel)
  EVT_LISTBOX(ID_TAS_RTCLICK_TTYPE, TASRightClickWin::OnSelection)
END_EVENT_TABLE()

TASRightClickWin::TASRightClickWin(TASTilePanel* parent)
                  : wxPanel(parent, wxID_ANY)
{
  m_pParent = parent;

  wxBoxSizer* sizer = new wxBoxSizer(wxVERTICAL);
  wxPanel* pPanel = new wxPanel(this);

  sizer->Add(pPanel);

  Hide();

  wxBoxSizer* pMainSizer = new wxBoxSizer(wxVERTICAL);
  m_pListBox = new wxListBox(pPanel, ID_TAS_RTCLICK_TTYPE);
  UpdateValues();
  pMainSizer->Add(m_pListBox);

  pPanel->SetAutoLayout(TRUE);
  pPanel->SetSizer(pMainSizer);
  
  pMainSizer->SetSizeHints(pPanel);
  pMainSizer->Fit(pPanel);

  SetAutoLayout(TRUE);
  SetSizer(sizer);
  
  sizer->SetSizeHints(this);
  sizer->Fit(this);
}

void TASRightClickWin::UpdateValues()
{
  if (m_pParent->IsInOverviewWin() || (g_pContext->m_iTASOptMode == TAS_OPT_MODE_0)) {
    return;
  }
  wxArrayString arr;
  arr.Insert(TAS_EMPTY_TTYPE_STR, 0);
  for (size_t i=0; i<g_pTileSpace->m_tileset.size(); i++) {
    arr.Insert(g_pTileSpace->m_tileset[i]->m_sName, i+1);
  }
  m_pListBox->Set(arr);
}

void TASRightClickWin::ShowEx(wxPoint& pt, TASCoords& coords)
{
  if (m_pParent->IsInOverviewWin() || (g_pContext->m_iTASOptMode == TAS_OPT_MODE_0)) {
    return;
  }

  m_coords = coords;

  //  Figure out what tiletype is currently at this location to preset the listbox selection
  TASTileType* pTT = g_pTileSpace->GetTileType(coords);
  if (pTT) {
    m_pListBox->SetStringSelection(pTT->m_sName);
  } else {
    m_pListBox->SetSelection(0);
  }

  Move(pt);

  //  Adjust window position in case it is close to the edge and not completely visible
  wxRect rectParent = m_pParent->GetRect();
  wxRect rectThis = GetRect();

  if ((rectThis.GetX() + rectThis.GetWidth()) > rectParent.GetWidth()) {
    int newX = pt.x - rectThis.GetWidth();
    if (newX > 0) {
      pt.x = newX;
    }
  }
  if ((rectThis.GetY() + rectThis.GetHeight()) > rectParent.GetHeight()) {
    int newY = pt.y - rectThis.GetHeight();
    if (newY > 0) {
      pt.y = newY;
    }
  }
  Move(pt);
  m_pParent->DrawLocation(m_coords, true);
  Show();
}

void TASRightClickWin::OnSelection(wxCommandEvent& event)
{
  wxArrayInt selections;
  int num = m_pListBox->GetSelections(selections);
  int ttnum = selections[0];
  Hide();

  if (g_pTileSpace->m_epoch > 0) {
    wxMessageBox(_T("The tile assembly can only be manually\nmodified in the seed state (i.e. before\nany simulation steps)"), _T("Unable to modify assembly"));
    return;
  }

  if (ttnum == 0) {     //  This was a removal (or setting to <empty>
    g_pTileSpace->RemoveTile(m_coords);
  } else {              //  An actual tile type was selected
    if (g_pTileSpace->GetTileType(m_coords)) {       //  There was already a tile here, so remove it first
      g_pTileSpace->RemoveTile(m_coords);
    }
    g_pTileSpace->AddTile(m_coords, g_pTileSpace->m_tileset[ttnum-1]);
  }
  m_pParent->DrawLocation(m_coords, true);
  TASCoords coordsFake;
  m_pParent->m_parentFrame->m_pCellValueWin->SetCurrentCell(coordsFake, false, false);
  m_pParent->m_parentFrame->m_pCellValueWin->SetCurrentCell(m_coords, false, false);
}

IMPLEMENT_CLASS(TASTilePanel, TASPanelBase)
BEGIN_EVENT_TABLE(TASTilePanel, TASPanelBase)
  EVT_PAINT(TASTilePanel::OnPaint)
  EVT_MOUSE_EVENTS(TASTilePanel::OnMouse)
  EVT_MOUSEWHEEL(TASTilePanel::OnMouseWheel)
  EVT_MENU(Menu_RightClick_TileAdd, TASTilePanel::OnRtClickMenu)
  EVT_KEY_DOWN(TASTilePanel::OnKeyboard)
  EVT_SCROLLWIN(TASTilePanel::OnScrollWin)
  EVT_SCROLL(TASTilePanel::OnScrollEx)
  EVT_SIZE(TASTilePanel::OnSize)
//  EVT_ERASE_BACKGROUND(TASTilePanel::OnEraseBackground)
END_EVENT_TABLE()

TASTilePanel::TASTilePanel(TASSimulatorFrame* parentFrame, TASTilePanel* pMainPanel, wxWindowID id,
                               const wxPoint& pos, const wxSize& size, long style, const wxString& name)
                 : TASPanelBase(parentFrame, id, pos, size, style, name)
{
  m_parentFrame     = parentFrame;
  m_pSpace          = NULL;
  m_lBeginX         = 0;
  m_lBeginY         = 0;
  m_fScaleFact      = 100.0f;
  m_bIsOverview     = false;
  m_lastDrawnIter   = 0;
  m_pRtClickWin     = NULL;
  m_bPanning        = false;
  m_drag_start_x    = 0;
  m_drag_start_y    = 0;
  m_rotationCount   = 0;
  m_zoomDirection   = 0;
  m_fZoomPercentage = 0.45f;
  m_zoomTicks       = 5;
  m_mouseMode       = TAS_MOUSE_MODE_DRAG;

  m_bCenterCursorOnDClick = false;

  m_brushOffspace   = wxSystemSettings::GetColour(wxSYS_COLOUR_APPWORKSPACE);
  m_penOffspace     = wxSystemSettings::GetColour(wxSYS_COLOUR_APPWORKSPACE);

  // Create a Selection Box (Rubber Band)
  m_RubberBand = wxRubberBand(this);
  m_rubberbandPen.SetStyle(wxSOLID);
  m_rubberbandPen.SetWidth(1);
  m_RubberBand.SetPen(m_rubberbandPen);

  m_rubberbandorigin = wxPoint(INVALID_LOC,INVALID_LOC);

  SetBackgroundStyle(wxBG_STYLE_CUSTOM);

  wxBitmap bm = wxXmlResource::Get()->LoadBitmap(_T("grab.gif"));
  if (bm.IsOk()) {
    m_cursorGrab      = bm.ConvertToImage();
  }
  wxBitmap bm2 = wxXmlResource::Get()->LoadBitmap(_T("grabbing.gif"));
  if (bm2.IsOk()) {
    m_cursorGrabbing  = bm2.ConvertToImage();
  }

  m_coordsCurrCell  = TASCoords(INT_MAX, INT_MAX, 0);

  if (pMainPanel == NULL) {
    SetMainPanel(this);
  } else {
    SetMainPanel(pMainPanel);
  }

  if (!IsInOverviewWin()) {
    m_pRtClickWin = new TASRightClickWin(this);
  }

  SetScrollRate(1, 1);
  EnableScrolling(false, false);

  DoCanvasResize();
}

TASTilePanel::~TASTilePanel()
{
}

bool TASTilePanel::IsValid() {
  return true;
}

void TASTilePanel::SetMainPanel(TASTilePanel* pPanel)
{
  m_pMainPanel = pPanel;
  if (this == m_pMainPanel) {
    m_bIsOverview = false;
  } else {
    m_bIsOverview = true;
  }
}

void TASTilePanel::DoScaling(wxDC* pDC)
{
  //  This is overridden for the overview window
  pDC->SetUserScale((GetScaleFact()/100), (GetScaleFact()/100));
}

void TASTilePanel::DrawLocation(TASCoords& coords, bool bNeighborhood)
{
  wxClientDC dc(this);
  PrepareForDrawing(&dc);
  
  int count = 1;
  if (bNeighborhood) {
    count = 5;
    if (g_pTileSpace->GetDimNum() == 3) {
      count = 7;
    }
  }

  for (int i=0; i<count; i++) {
    TASCoords coordsCurr = coords + coordsNbrhd[i];
    TASTileType* pTT = g_pTileSpace->GetTileType(coordsCurr);
    bool bCurrent = (coordsCurr == m_coordsCurrCell);
    if (pTT) {
      DrawTileEx(&dc, coordsCurr, pTT, TAS_DRAW_TILE, bCurrent);
    } else if (m_bShowFrontier && g_pTileSpace->IsFrontierLocation(coordsCurr)) {
      if (g_pTileSpace->IsDeadFrontierLocation(coordsCurr)) {
        DrawTileEx(&dc, coordsCurr, NULL, TAS_DRAW_FRONTIERDEAD);
      } else {
        DrawTileEx(&dc, coordsCurr, NULL, TAS_DRAW_FRONTIER);
      }
    } else {
      DrawTileEx(&dc, coordsCurr, NULL, TAS_DRAW_EMPTY, bCurrent);
    }
  }
}

void TASTilePanel::DrawTileEx(wxDC* pDC, TASCoords& coords, TASTileType* pTT, int drawType, bool bOutline)
{
  if (drawType == TAS_DRAW_INVALID) {
    return;
  }

  wxPoint pt;
  GetTileLocation(coords, &pt.x, &pt.y);

  if (IsInOverviewWin()) {
    DoScaling(pDC);
  }

  if (drawType == TAS_DRAW_FRONTIER) {
    m_ttFrontierTile.DrawTile(pDC, pt, m_tileSize, 0);
  } else if (drawType == TAS_DRAW_FRONTIERDEAD) {
    m_ttFrontierDeadTile.DrawTile(pDC, pt, m_tileSize, 0);
  } else if ((drawType == TAS_DRAW_EMPTY) || !pTT) {
//MJP: turn this into a tiletype(?)
    pDC->SetPen(m_penBackground);
    pDC->SetBrush(m_brushBackground);
    pDC->DrawRectangle(pt.x, pt.y, m_tileSize, m_tileSize);
  } else {
    pTT->DrawTile(pDC, pt, m_tileSize, g_pContext->m_maskSimTileDisplay);
  }

  if (bOutline) {
    pDC->SetPen(*wxRED_PEN);
    pDC->SetBrush(*wxTRANSPARENT_BRUSH);
    pDC->DrawRectangle(pt.x, pt.y, m_tileSize, m_tileSize);
  }
}

//size_t overviewRedrawCounter = 0;

bool TASTilePanel::RenderTileSpace(wxDC* pDC)
{
  //  This function should currently only be used for drawing the overview window
  //  since it completely ignores the frontiers
  wxASSERT(IsInOverviewWin());

/*
overviewRedrawCounter++;
wxString dbgMsg = wxString::Format("OverviewWin redraw #%d\n", overviewRedrawCounter);
g_pTASApp->SetSimDebugText(dbgMsg, true);
*/

  size_t t = 0;
  TASLocationTransition* pTrans = NULL;

  int typeFrontier = TAS_DRAW_FRONTIER;
  int typeDeadFrontier = TAS_DRAW_FRONTIERDEAD;
  if (!m_bShowFrontier || IsInOverviewWin()) {
    typeFrontier = TAS_DRAW_EMPTY;
    typeDeadFrontier = TAS_DRAW_EMPTY;
  }

//MJP: may want to rework this to iterate over the entire space (on full frontier addition?)
  for (size_t e=0; e<=(size_t)g_pTileSpace->m_epoch; e++) {
    t = 0;
    while (pTrans = g_pTileSpace->GetTransitionToTypeInEpoch(e, t++, TAS_BEGIN_TILES)) {
/*
      int type = TAS_DRAW_EMPTY;
      if (TASIsTileType(pTrans->m_typeEnd)) {
        type = TAS_DRAW_TILE;
      } else if (pTrans->m_typeEnd == TAS_TILE_FRONTIER) {
        type = typeFrontier;
      } else if (pTrans->m_typeEnd == TAS_TILE_FRONTIERDEAD) {
        type = typeDeadFrontier;
      }
*/
      int type = TAS_DRAW_TILE;

      if (m_pMainPanel->m_iViewBy == TAS_3D_VIEWBY_UNLIMITED) {
        DrawTileEx(pDC, pTrans->m_coords, g_pTileSpace->m_tileset[pTrans->m_typeEnd], type);
      } else if (m_pMainPanel->m_iViewBy == TAS_3D_VIEWBY_RANGE) {
        if (pTrans->m_coords.x >= m_pMainPanel->m_coordsMinVis.x && pTrans->m_coords.x <= m_pMainPanel->m_coordsMaxVis.x &&
            pTrans->m_coords.y >= m_pMainPanel->m_coordsMinVis.y && pTrans->m_coords.y <= m_pMainPanel->m_coordsMaxVis.y &&
            pTrans->m_coords.z >= m_pMainPanel->m_coordsMinVis.z && pTrans->m_coords.z <= m_pMainPanel->m_coordsMaxVis.z)
        {
          DrawTileEx(pDC, pTrans->m_coords, g_pTileSpace->m_tileset[pTrans->m_typeEnd], type);
        }
      } else if (m_pMainPanel->m_iViewBy == TAS_3D_VIEWBY_HOOD) {
        if (m_pMainPanel->CoordsInNeighborhood(pTrans->m_coords))
        {
          DrawTileEx(pDC, pTrans->m_coords, g_pTileSpace->m_tileset[pTrans->m_typeEnd], type);
        }
      }
    }
  }
  return true;
}

void TASTilePanel::DrawIteration(int epoch)
{
  //  Handle case where appearance isn't refreshed every iteration - just do full redraw
/*  if (abs((int)(epoch - m_lastDrawnIter)) > 1) {
    UpdateSpace(true);
    return;
  }*/

  wxMutexLocker lock(*g_pMutexTileSpaceGuard);

  CheckCanvasSize();

//MJP: this could be either further optimized or abstracted to avoid duplication with ::OnPaint flow
  bool bReverse = false;
  if (epoch < m_lastDrawnIter) {
    epoch++;
    bReverse = true;
  }

  int typeFrontier = TAS_DRAW_FRONTIER;
  int typeDeadFrontier = TAS_DRAW_FRONTIERDEAD;
  if (!m_bShowFrontier || IsInOverviewWin()) {
    typeFrontier = TAS_DRAW_EMPTY;
    typeDeadFrontier = TAS_DRAW_EMPTY;
  }

  size_t numChanges = g_pTileSpace->GetNumTransitionsInEpoch(epoch);
  if (numChanges > 0) {
    wxClientDC dc(this);
    PrepareForDrawing(&dc);

    for (size_t i=0; i<numChanges; i++) {

      TASLocationTransition* pTrans = g_pTileSpace->GetTransitionInEpoch(epoch, i);
      if (pTrans) {

        size_t typeResult = pTrans->m_typeEnd;
        if (bReverse) {
          typeResult = pTrans->m_typeBegin;
        }

        int drawType = TAS_DRAW_EMPTY;
        TASTileType* pTT  = NULL;
        if (TASIsTileType(typeResult)) {
          drawType = TAS_DRAW_TILE;
          pTT = g_pTileSpace->m_tileset[typeResult];
        } else if (typeResult == TAS_TILE_FRONTIER) {
          drawType = typeFrontier;
        } else if (typeResult == TAS_TILE_FRONTIERDEAD) {
          drawType = typeDeadFrontier;
        }

        if (m_pMainPanel->m_iViewBy == TAS_3D_VIEWBY_UNLIMITED) {
          DrawTileEx(&dc, pTrans->m_coords, pTT, drawType);
        } else if (m_pMainPanel->m_iViewBy == TAS_3D_VIEWBY_RANGE) {
          if (pTrans->m_coords.x >= m_pMainPanel->m_coordsMinVis.x && pTrans->m_coords.x <= m_pMainPanel->m_coordsMaxVis.x &&
              pTrans->m_coords.y >= m_pMainPanel->m_coordsMinVis.y && pTrans->m_coords.y <= m_pMainPanel->m_coordsMaxVis.y &&
              pTrans->m_coords.z >= m_pMainPanel->m_coordsMinVis.z && pTrans->m_coords.z <= m_pMainPanel->m_coordsMaxVis.z)
          {
            DrawTileEx(&dc, pTrans->m_coords, pTT, drawType);
          }
        } else if (m_pMainPanel->m_iViewBy == TAS_3D_VIEWBY_HOOD) {
          if (m_pMainPanel->CoordsInNeighborhood(pTrans->m_coords))
          {
            DrawTileEx(&dc, pTrans->m_coords, pTT, drawType);
          }
        }
      }
    }

    if (IsInOverviewWin()) {
      dc.SetUserScale(1.0, 1.0);
      DrawOverlay(&dc, false);
    }
  }

  m_lastDrawnIter = epoch;
}

void TASTilePanel::DrawOverlay(wxDC* WXUNUSED(pDC), bool WXUNUSED(bOnlyDraw))
{
  return;
}

void TASTilePanel::CheckCanvasSize()
{
  if ((m_pMainPanel->m_coordsMinVis != g_pTileSpace->m_coordsMin) ||
      (m_pMainPanel->m_coordsMaxVis != g_pTileSpace->m_coordsMax))
  {
    DoCanvasResize();
  }
}

void TASTilePanel::OnEraseBackground(wxEraseEvent& event)
{

}

void TASTilePanel::PrepareForDrawing(wxDC* pDC)
{
  DoPrepareDC(*pDC);

  SetBackgroundColour(g_pContext->m_colSpaceBkgrnd);
  m_brushBackground.SetColour(g_pContext->m_colSpaceBkgrnd);
  m_penBackground.SetColour(g_pContext->m_colSpaceBkgrnd);

  pDC->SetBackground(m_brushBackground);

  pDC->SetFont(*wxSWISS_FONT);
  pDC->SetTextForeground(g_pContext->m_colDefaultText);

  m_brush.SetColour(g_pContext->m_colDefaultTile);
  m_pen.SetColour(g_pContext->m_colDefaultText);

  pDC->SetBrush(m_brush);
  pDC->SetPen(m_pen);
}

void TASTilePanel::OnPaint(wxPaintEvent& WXUNUSED(event))
{
  if (!m_pMainPanel) return;

  wxMutexLocker lock(*g_pMutexTileSpaceGuard);

  CheckCanvasSize();

  wxAutoBufferedPaintDC dcAutoBuff(this);
  PrepareForDrawing(&dcAutoBuff);

  dcAutoBuff.SetBackground(m_brushOffspace);
  dcAutoBuff.Clear();

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*
    int xViewStart = 0;
    int yViewStart = 0;
    GetViewStart(&xViewStart, &yViewStart);

    int xScrollPixelsPerUnit = 0;
    int yScrollPixelsPerUnit = 0;
    GetScrollPixelsPerUnit(&xScrollPixelsPerUnit, &yScrollPixelsPerUnit);

    int xClientWidth  = 0;
    int yClientHeight = 0;
    GetClientSize(&xClientWidth, &yClientHeight);

    int virtWidth  = 0;
    int virtHeight = 0;
    GetVirtualSize(&virtWidth, &virtHeight);

    dcAutoBuff.DestroyClippingRegion();

    //dc.SetClippingRegion((xViewStart * xScrollPixelsPerUnit), (yViewStart * yScrollPixelsPerUnit),
    //                     xClientWidth, yClientHeight);
    dcAutoBuff.SetClippingRegion(0, 0, virtWidth, virtHeight);
*/
/////////////////////////////////////////////////////////////////////////////////////////////////////////////

//  Don't do this!  BAD!!  ClearBackground();

///////////////////////////////////
//MJP: the following block could be filled out to further optimize redraw to only 'damaged' area
//MJP: if update region is close to entire space, do regular full space drawing
  if (!IsInOverviewWin()) {
    // Find Out where the window is scrolled to
    int vbX,vbY;                              // Top left corner of client
    GetViewStart(&vbX,&vbY);

    TASCoords coordsCurr;
    int vX,vY,vW,vH;                          // Dimensions of client area in pixels
    wxRegionIterator upd(GetUpdateRegion());  // get the update rect list

    TASTileType* pTT = NULL;

    while (upd)
    {
      vX = vbX + upd.GetX();
      vY = vbY + upd.GetY();
      vW = upd.GetW();
      vH = upd.GetH();

      // Repaint this rectangle

      dcAutoBuff.SetBrush(m_brushOffspace);
      dcAutoBuff.SetPen(m_penOffspace);
      dcAutoBuff.DrawRectangle(vX, vY, vW, vH);

      int xCurr = vX;
      int xSave = xCurr;
      int yCurr = vY;
      int xMax  = vX + vW + m_tileSize;
      int yMax  = vY + vH + m_tileSize;

      // Find the tiles in this rectangle
      while (yCurr < yMax) {
        xCurr = xSave;
        while (xCurr < xMax) {
          GetCoordsFromLocation(xCurr, yCurr, &coordsCurr);
          if (pTT = g_pTileSpace->GetTileType(coordsCurr)) {
            DrawTileEx(&dcAutoBuff, coordsCurr, pTT, TAS_DRAW_TILE);
//These next three should be optimized!!!
          } else if (m_bShowFrontier && g_pTileSpace->IsDeadFrontierLocation(coordsCurr)) {
            DrawTileEx(&dcAutoBuff, coordsCurr, NULL, TAS_DRAW_FRONTIERDEAD);
          } else if (m_bShowFrontier && g_pTileSpace->IsFrontierLocation(coordsCurr)) {
            DrawTileEx(&dcAutoBuff, coordsCurr, NULL, TAS_DRAW_FRONTIER);
          } else if (coordsCurr.IsValid()) {
            DrawTileEx(&dcAutoBuff, coordsCurr, NULL, TAS_DRAW_EMPTY);
          }
          xCurr += m_tileSize;
        }
        yCurr += m_tileSize;
      }

      upd++;
    }
  } else {

    dcAutoBuff.SetBackground(m_brushBackground);
    dcAutoBuff.Clear();

    RenderTileSpace(&dcAutoBuff);

    dcAutoBuff.SetUserScale(1.0, 1.0);

    DrawOverlay(&dcAutoBuff, true);
  }

  m_lastDrawnIter = g_pTileSpace->m_epoch;
}

void TASTilePanel::Reset()
{
  SetScaleFact(100);
}

void TASTilePanel::DoCanvasResize(bool bNoRefresh, bool bRefreshOverview)
{
  if (m_parentFrame->m_pOverviewWin) {
    ((TASTileOverviewWin*)m_parentFrame->m_pOverviewWin)->ClearOverlay(NULL);
  }

  if (!m_bIsOverview) {

    m_tileSize = (g_tileSize * (m_fScaleFact / 100));

    //  Figure out what tile space location the viewable portion of the screen is currently centered on
    int xCenterPixel = 0;
    int yCenterPixel = 0;
    GetCenterViewablePixel(&xCenterPixel, &yCenterPixel, false);

    TASCoords coordsCenter;
    GetCoordsFromLocation(xCenterPixel, yCenterPixel, &coordsCenter);

    //  Calculate the necessary screen size
    if (m_pMainPanel->m_iViewBy == TAS_3D_VIEWBY_UNLIMITED) {
      m_pMainPanel->m_coordsMinVis = g_pTileSpace->m_coordsMin;
      m_pMainPanel->m_coordsMaxVis = g_pTileSpace->m_coordsMax;
    }

//MJP: need to handle other viewby possibilities!!
    m_virtWidth  = (m_coordsMaxVis.x - m_coordsMinVis.x + 1) * m_tileSize;
    m_virtHeight = (m_coordsMaxVis.y - m_coordsMinVis.y + 1) * m_tileSize;

    m_coordsCurrCell.Invalidate();

    if (m_parentFrame->m_pOverviewWin) {
      ((TASTileOverviewWin*)m_parentFrame->m_pOverviewWin)->m_tileSize = m_tileSize;
    }

    wxString dbgMsg = wxString::Format(_T("Setting canvas size to %d x %d for tilesize %d and dims (%d, %d) x (%d, %d)\n"),
                                       m_virtWidth, m_virtHeight,
                                       m_tileSize,
                                       m_coordsMinVis.x, m_coordsMinVis.y,
                                       m_coordsMaxVis.x, m_coordsMaxVis.y);
    g_pTASApp->SetSimDebugText(dbgMsg, true);

    //  Resize the window
    SetVirtualSize(m_virtWidth, m_virtHeight);

    //  Scroll back to the location that was centered before resizing
    ScrollToCenter(coordsCenter, true);

    //SetScrollRate((g_tileSize * 10), (g_tileSize * 10));
    //SetScrollRate(100, 100);

    if (!bNoRefresh) {
      Refresh();
    }

//    m_bitmap.SetSize(m_virtWidth, m_virtHeight);
//    m_bitmap = m_bitmap.GetSubBitmap(wxRect(0, 0, m_bitmap.GetWidth(), m_bitmap.GetHeight()));

    if (bRefreshOverview) {
      m_parentFrame->RefreshOverviewWin();
    }
  }
}

float TASTilePanel::SetScaleFact(float fScaleFact, TASCoords* pCtrCoords, bool bKeepRelativePosition)
{
  SetCursor(*wxHOURGLASS_CURSOR);

  TASCoords coordsCenter;
  wxPoint ptStart;
  int xVSOffset = 0;
  int yVSOffset = 0;
  if (pCtrCoords && !bKeepRelativePosition) {
    coordsCenter = *pCtrCoords;
  } else if (pCtrCoords) {
    GetTileLocation(*pCtrCoords, &ptStart.x, &ptStart.y);
    int xVS = 0;
    int yVS = 0;
    GetViewStart(&xVS, &yVS);
    xVSOffset = ptStart.x - xVS;
    yVSOffset = ptStart.y - yVS;
  } else {
    int xCenterPixel = 0;
    int yCenterPixel = 0;
    GetCenterViewablePixel(&xCenterPixel, &yCenterPixel, true);
    GetCoordsFromLocation(xCenterPixel, yCenterPixel, &coordsCenter, true);
  }

  m_fScaleFact = fScaleFact;

  //  Don't allow scaling to become too small
  if (m_fScaleFact < 1.0f) {
    m_fScaleFact = 1.0f;
  }

  //  Don't allow scaling to become too large
  if (m_fScaleFact > 1000.0f) {
    m_fScaleFact = 1000.0f;
  }

  m_tileSize = (g_tileSize * (m_fScaleFact / 100));
  if (m_parentFrame->m_pOverviewWin) {
    ((TASTileOverviewWin*)m_parentFrame->m_pOverviewWin)->m_tileSize = m_tileSize;
  }

  m_parentFrame->SetZoomLevel(m_fScaleFact);

  DoCanvasResize(true, false);

  //  Attempt to scroll back to same position
  if (!bKeepRelativePosition) {
    ScrollToCenter(coordsCenter, false);
  } else {
    wxPoint ptNew;
    GetTileLocation(*pCtrCoords, &ptNew.x, &ptNew.y);

    int xScrollPixelsPerUnit = 0;
    int yScrollPixelsPerUnit = 0;
    GetScrollPixelsPerUnit(&xScrollPixelsPerUnit, &yScrollPixelsPerUnit);

    int xNew = ptNew.x - xVSOffset;
    int yNew = ptNew.y - yVSOffset;

    Scroll((xNew / xScrollPixelsPerUnit), (yNew / yScrollPixelsPerUnit));
  }

  Refresh();

  SetCursor(wxNullCursor);

  return m_fScaleFact;
}

float TASTilePanel::GetScaleFact()
{
  return m_fScaleFact;
}

void TASTilePanel::GetCenterViewablePixel(int *x, int *y, bool bFromScaling)
{
    int xViewStart = 0;
    int yViewStart = 0;
    GetViewStart(&xViewStart, &yViewStart);

    int xScrollPixelsPerUnit = 0;
    int yScrollPixelsPerUnit = 0;
    GetScrollPixelsPerUnit(&xScrollPixelsPerUnit, &yScrollPixelsPerUnit);

    int xClientWidth  = 0;
    int yClientHeight = 0;
    GetClientSize(&xClientWidth, &yClientHeight);

    int virtWidth  = 0;
    int virtHeight = 0;
    GetVirtualSize(&virtWidth, &virtHeight);

    //  Check to see if the tile space size is equal to or smaller than
    //  the viewable area, and if so, return the center of the tile space
    if ((xClientWidth >= virtWidth) || (yClientHeight >= virtHeight)) {
      *x = (g_pTileSpace->m_coordsDims.x * m_tileSize) / 2;
      *y = (g_pTileSpace->m_coordsDims.y * m_tileSize) / 2;
    } else {
      *x = (xViewStart * xScrollPixelsPerUnit) + (xClientWidth  / 2);
      if (bFromScaling) {
        *y = (yViewStart * yScrollPixelsPerUnit) + (yClientHeight / 2);
      } else {
        *y = (yViewStart * yScrollPixelsPerUnit) - (yClientHeight / 2);
      }
    }

    return;
}

void TASTilePanel::GetTileLocation(TASCoords& coords, int* x, int* y)
{
  //  Return the top left corner of the tile location
  *x = (coords.x - g_pTileSpace->m_coordsMin.x) * m_tileSize;
  *y = (g_pTileSpace->m_coordsMax.y - coords.y) * m_tileSize;
}

void TASTilePanel::GetCoordsFromLocation(int x, int y, TASCoords* coords, bool bUseTilespaceDims)
{
  if (bUseTilespaceDims) {
    coords->x = (x / m_tileSize) + g_pTileSpace->m_coordsMin.x;
    coords->y = g_pTileSpace->m_coordsMax.y - (y / m_tileSize);
    coords->z = 0;
  } else {
    coords->x = (x / m_tileSize) + m_coordsMinVis.x;
    coords->y = m_coordsMaxVis.y - (y / m_tileSize);
    coords->z = 0;
  }

  //  Make sure location is valid
  if (coords->x < g_pTileSpace->m_coordsMin.x || coords->x > g_pTileSpace->m_coordsMax.x ||
      coords->y < g_pTileSpace->m_coordsMin.y || coords->y > g_pTileSpace->m_coordsMax.y)
  {
    coords->Invalidate();
  }
}

void TASTilePanel::Scroll(int x, int y)
{
/*
  Update();
  if (!m_bPanning) {
    SetScrollRate(100, 100);
  } else {
    SetScrollRate(1, 1);
  }
*/
  wxScrollHelper::Scroll(x, y);
  if (!m_bIsOverview && m_parentFrame->m_pOverviewWin) {
    ((TASTileOverviewWin*)m_parentFrame->m_pOverviewWin)->DrawOverlay(NULL, false);
  }
}

void TASTilePanel::OnMouseWheel(wxMouseEvent& event)
{
  int wheelRotation = event.GetWheelRotation();

  if (wheelRotation != 0) {
    
    float newScale = GetScaleFact() + (GetScaleFact() * ((wheelRotation/120) * m_fZoomPercentage));//- m_fZoomIncrement;

    //  Figure out the coordinates that the cursor is currently over
    wxClientDC dc(this);
    PrepareForDrawing(&dc);
    wxPoint pt = event.GetLogicalPosition(dc);
    int winX, winY;
    GetPosition(&winX, &winY);
    long x = pt.x - winX;
    long y = pt.y - winY;
    TASCoords coordsCurrCell;
    GetCoordsFromLocation(x, y, &coordsCurrCell, true);

    SetScaleFact(newScale, &coordsCurrCell, true);
  }
}

void TASTilePanel::OnFrontierMouse(wxMouseEvent& event, TASCoords& coordsCurr)
{
  if (event.Leaving()) {

    m_coordsCurrCell.Invalidate();
    m_parentFrame->UpdateStatusBar();

    SetCursor(wxNullCursor);

    if (m_parentFrame->m_pCellValueWin) {
      m_parentFrame->m_pCellValueWin->SetCurrentCell(m_pMainPanel->m_coordsCurrCell, false, false);
    }

    m_RubberBand.Erase();
    ResetRubberband();
    return;
  }

  m_coordsCurrCell = coordsCurr;

  if (m_parentFrame->m_pCellValueWin) {
    m_parentFrame->m_pCellValueWin->SetCurrentCell(m_coordsCurrCell, false, false);
  }

  m_parentFrame->UpdateStatusBar();
  long i = coordsCurr.x;
  long j = coordsCurr.y;

  // NOTE that wxMouseEvent::LeftDown() and wxMouseEvent::LeftIsDown()
  // have different semantics. The first one is used to signal that the
  // button was just pressed (i.e., in "button down" events); the second
  // one just describes the current status of the button, independently
  // of the mouse event type. LeftIsDown is typically used in "mouse
  // move" events, to test if the button is _still_ pressed.

  if (event.LeftUp() || event.RightUp()) {
    wxRect r = m_RubberBand.GetRect();

    int startX, startY;
    GetViewStart(&startX, &startY);

    ResetRubberband();
    SetCursor(wxNullCursor);

    TASCoords coordsTL;
    GetCoordsFromLocation(r.GetLeft()+startX, r.GetTop()+startY, &coordsTL);
    coordsTL.x++;
    coordsTL.y--;

    TASCoords coordsBR;
    GetCoordsFromLocation(r.GetRight()+startX, r.GetBottom()+startY, &coordsBR);
    coordsBR.x--;
    coordsBR.y++;

    wxString sMsg;

    if (event.LeftUp()) {
      TASFrontierSet::iterator it = g_pTileSpace->m_setFrontier.begin();
      while (it != g_pTileSpace->m_setFrontier.end())
      {
        if ((it->m_coords.x >= coordsTL.x) && (it->m_coords.x <= coordsBR.x) &&
            (it->m_coords.y <= coordsTL.y) && (it->m_coords.y >= coordsBR.y))
        {
          size_t indexLoc = g_pTileSpace->Index(it->m_coords);

          if (g_pTileSpace->IsFrontierLocation(it->m_coords) &&
              !g_pTileSpace->IsDeadFrontierLocation(it->m_coords))
          {
            g_pTileSpace->m_pLocationIndex->SetLocation(indexLoc, TAS_TILE_FRONTIERDEAD, g_pTileSpace->m_epoch);
            TASFrontierLocation tfl(it->m_coords);
            g_pTileSpace->m_setFrontierDead.insert(tfl);
            g_pTileSpace->m_setFrontier.erase(it++);
            sMsg += tfl.m_coords.ToString() + _T(" ");
          } else {
            it++;
          }
        } else {
          it++;
        }
      }
    } else if (event.RightUp()) {
      TASFrontierSet::iterator it = g_pTileSpace->m_setFrontierDead.begin();
      while (it != g_pTileSpace->m_setFrontierDead.end())
      {
        if ((it->m_coords.x >= coordsTL.x) && (it->m_coords.x <= coordsBR.x) &&
            (it->m_coords.y <= coordsTL.y) && (it->m_coords.y >= coordsBR.y))
        {
          size_t indexLoc = g_pTileSpace->Index(it->m_coords);

          if (g_pTileSpace->IsDeadFrontierLocation(it->m_coords)) {
            g_pTileSpace->m_pLocationIndex->SetLocation(indexLoc, TAS_TILE_FRONTIER, g_pTileSpace->m_epoch);
            TASFrontierLocation tfl(it->m_coords);
            g_pTileSpace->m_setFrontier.insert(tfl);
            g_pTileSpace->m_setFrontierDead.erase(it++);
            sMsg += tfl.m_coords.ToString() + _T(" ");
          } else {
            it++;
          }
        } else {
          it++;
        }
      }
    }
    Refresh();
    if (sMsg.Len() > 0) {   //  This means that there was some change to the frontier
      //  Truncate the epoch cache because anything from future epochs is now invalid
      g_pTileSpace->TruncateEpochCache(g_pTileSpace->m_epoch);

      //  Display a message about what was changed (if debug messages are turned on)
      sMsg = _T("Frontier locations ") + sMsg + _T(" were turned ");
      if (event.RightUp()) {
        sMsg += _T("on.\n");
      } else {
        sMsg += _T("off.\n");
      }
      g_pTASApp->SetSimDebugText(sMsg, true);
    }
  } else if (!event.LeftIsDown() && !event.RightIsDown()) {
//      m_status = MOUSE_NOACTION;
      return;
  } else if (event.LeftDown() || event.RightDown()) {
    m_rubberbandorigin = event.GetPosition();
    m_lBeginX = i;
    m_lBeginY = j;
    return;
  } else if (((m_lBeginX != i) || (m_lBeginY != j)) &&
             (m_lBeginX != INVALID_LOC) && (m_lBeginY != INVALID_LOC))
  {
    wxPoint myActualPosition = event.GetPosition();
    int xIn, yIn, wIn, hIn;
    if (myActualPosition.x < m_rubberbandorigin.x) {
      xIn = myActualPosition.x;
      wIn = m_rubberbandorigin.x - myActualPosition.x;
    } else {
      xIn = m_rubberbandorigin.x;
      wIn = myActualPosition.x - m_rubberbandorigin.x;
    }
    if (myActualPosition.y < m_rubberbandorigin.y) {
      yIn = myActualPosition.y;
      hIn = m_rubberbandorigin.y - myActualPosition.y;
    } else {
      yIn = m_rubberbandorigin.y;
      hIn = myActualPosition.y - m_rubberbandorigin.y;
    }
    m_RubberBand.SetGeometry(xIn, yIn, wIn, hIn);
    return;
  }
}

void TASTilePanel::UnselectCell(wxDC* pDC)
{
  if (m_coordsSelCell.IsValid()) {
    if (m_bShowFrontier && (g_pTileSpace->IsFrontierLocation(m_coordsSelCell))) {
      if (g_pTileSpace->IsDeadFrontierLocation(m_coordsSelCell)) {
        DrawTileEx(pDC, m_coordsSelCell, NULL, TAS_DRAW_FRONTIERDEAD);
      } else {
        DrawTileEx(pDC, m_coordsSelCell, NULL, TAS_DRAW_FRONTIER);
      }
    } else {
      DrawTileEx(pDC, m_coordsSelCell, g_pTileSpace->GetTileType(m_coordsSelCell), TAS_DRAW_TILE);
    }
    m_coordsSelCell.Invalidate();
  }
}

void TASTilePanel::OnMouse(wxMouseEvent& event)
{
  wxClientDC dc(this);

  PrepareForDrawing(&dc);

  //  Figure out the coordinates that the cursor is currently over
  m_ptMouse = event.GetLogicalPosition(dc);
  long x = m_ptMouse.x;
  long y = m_ptMouse.y;
  TASCoords coordsCurrCell;
  GetCoordsFromLocation(x, y, &coordsCurrCell, true);

  if (m_mouseMode == TAS_MOUSE_MODE_FRONTIER) {
    OnFrontierMouse(event, coordsCurrCell);
    return;
  }

  //////////////////////////////////////////
  //  Handling for mouse-based zooming
  if (coordsCurrCell.IsValid()) {
    int wheelRotation = event.GetWheelRotation();
    if (event.MiddleDown()) {
      m_ptZooming = event.GetPosition();
    } else if ((wheelRotation != 0) || (event.MiddleIsDown() && event.Dragging())) {
      bool bFirst = (m_rotationCount == 0);
      if (wheelRotation > 0) {
        m_rotationCount = (m_rotationCount + wheelRotation) % m_zoomTicks;
        m_zoomDirection = TAS_ZOOM_IN;
      } else if (wheelRotation < 0) {
        m_rotationCount = (m_rotationCount - wheelRotation) % m_zoomTicks;
        m_zoomDirection = TAS_ZOOM_OUT;
      } else {
        m_rotationCount = (m_rotationCount + 1) % m_zoomTicks;
        if (bFirst) {
          if (event.GetY() < m_ptZooming.y) {
            m_zoomDirection = TAS_ZOOM_IN;
          } else if (event.GetY() > m_ptZooming.y) {
            m_zoomDirection = TAS_ZOOM_OUT;
          }
        }
      }

      if (!bFirst) {
        WarpPointer(m_ptZooming.x, m_ptZooming.y);
      }
      if (m_rotationCount == (m_zoomTicks - 1)) {
        wxPoint currPt = event.GetLogicalPosition(dc);
        TASCoords coordsCurr;
        GetCoordsFromLocation(currPt.x, currPt.y, &coordsCurr);
        if (m_zoomDirection == TAS_ZOOM_IN) {           //  Increase zoom level
          float newScale = GetScaleFact() + (GetScaleFact() * m_fZoomPercentage);// + m_fZoomIncrement;
          SetScaleFact(newScale, &coordsCurr, true);
          return;
        } else if (m_zoomDirection == TAS_ZOOM_OUT) {   //  Decrease zoom level
          float newScale = GetScaleFact() - (GetScaleFact() * m_fZoomPercentage);//- m_fZoomIncrement;
          SetScaleFact(newScale, &coordsCurr, true);
          return;
        }
      }
    } else if (event.MiddleUp()) {
      m_rotationCount = 0;
      m_zoomDirection = 0;
    }
  }
  //////////////////////////////////////////

  if (m_bPanning) {
    SetCursor(m_cursorGrabbing);
  } else if (coordsCurrCell.IsValid()) {
    SetCursor(m_cursorGrab);
  } else {
    SetCursor(wxNullCursor);
  }

  if (m_pRtClickWin && m_pRtClickWin->IsShown() && event.ButtonDown()) {
    m_pRtClickWin->Hide();
  }

  if (event.LeftDClick()) {
    if (coordsCurrCell.IsValid()) {
      m_coordsCurrCell = coordsCurrCell;
      ScrollToCenter(m_coordsCurrCell);

      if (m_bCenterCursorOnDClick) {
        int iX = 0;
        int iY = 0;
        GetTileLocation(m_coordsCurrCell, &iX, &iY);

        int xVS, yVS;
        GetViewStart(&xVS, &yVS);
        iX = iX - xVS + (m_tileSize / 2);
        iY = iY - yVS + (m_tileSize / 2);
        WarpPointer(iX, iY);
      }
      return;
    }
  } else if (event.ButtonDown() && event.LeftIsDown()) {

    if (m_coordsSelCell != coordsCurrCell) {
      UnselectCell(&dc);
    }
    m_coordsSelCell = coordsCurrCell;

		int scroll_x, scroll_y;

//      int xSR, ySR;
//      GetScrollPixelsPerUnit(&xSR, &ySR);
//            SetScrollRate(1, 1);

    GetViewStart(&scroll_x, &scroll_y);
    // start panning
    m_bPanning = true;
    m_drag_start_x = event.GetPosition().x + scroll_x;
    m_drag_start_y = event.GetPosition().y + scroll_y;

    SetCursor(m_cursorGrabbing);

//      SetScrollRate(xSR, ySR);
  }

  if (event.Dragging()) {
    UnselectCell(&dc);

//    wxSize sz(GetClientSize());

    if (m_bPanning) {

//      int xSR, ySR;
//      GetScrollPixelsPerUnit(&xSR, &ySR);
//      SetScrollRate(1, 1);
      // pan
      int dx = m_drag_start_x - event.GetPosition().x;
      int dy = m_drag_start_y - event.GetPosition().y;

      Scroll(dx, dy);

    }/* else if (event.RightIsDown()) {
      float y1 = (sz.y - 2.0 * m_lBeginY) / sz.y;
      float y2 = (sz.y - 2.0 * event.GetY()) / sz.y;
      float diff = y1 - y2;

      if (diff > 5.0f) {
        SetCursor(wxCURSOR_MAGNIFIER);
        SetScaleFact(m_fScaleFact + diff*15);
      }
    }*/
  } else if (event.LeftUp()) {
    m_bPanning = false;
    //Refresh();
  } else if (event.RightIsDown() && !IsInOverviewWin()) {

    wxPoint pt = event.GetPosition();
    if (m_pRtClickWin) {
      m_pRtClickWin->ShowEx(pt, m_coordsCurrCell);
    }

  } else if (event.GetEventType() == wxEVT_LEAVE_WINDOW) {
    SetCursor(wxNullCursor);
    if (m_bShowMouseOver && m_coordsCurrCell.IsValid()) {
      if (m_bShowFrontier && (g_pTileSpace->IsFrontierLocation(m_coordsCurrCell))) {
        if (g_pTileSpace->IsDeadFrontierLocation(m_coordsCurrCell)) {
          DrawTileEx(&dc, m_coordsCurrCell, NULL, TAS_DRAW_FRONTIERDEAD);
        } else {
          DrawTileEx(&dc, m_coordsCurrCell, NULL, TAS_DRAW_FRONTIER);
        }
      } else {
        DrawTileEx(&dc, m_coordsCurrCell, g_pTileSpace->GetTileType(m_coordsCurrCell), TAS_DRAW_TILE);
      }
    }
    m_coordsCurrCell.Invalidate();

    if (m_parentFrame->m_pCellValueWin) {
      if (!m_coordsSelCell.IsValid()) {
        m_parentFrame->m_pCellValueWin->SetCurrentCell(m_pMainPanel->m_coordsCurrCell, false, false);
      }
    }
  } else if (m_bShowMouseOver && (!m_pRtClickWin || !m_pRtClickWin->IsShown())) {
//    SetCursor(wxNullCursor);
    if (m_coordsCurrCell.IsValid()) {
      if (m_bShowFrontier && (g_pTileSpace->IsFrontierLocation(m_coordsCurrCell))) {
        if (g_pTileSpace->IsDeadFrontierLocation(m_coordsCurrCell)) {
          DrawTileEx(&dc, m_coordsCurrCell, NULL, TAS_DRAW_FRONTIERDEAD);
        } else {
          DrawTileEx(&dc, m_coordsCurrCell, NULL, TAS_DRAW_FRONTIER);
        }
      } else {
        DrawTileEx(&dc, m_coordsCurrCell, g_pTileSpace->GetTileType(m_coordsCurrCell), TAS_DRAW_TILE);
      }
    }

    m_coordsCurrCell = coordsCurrCell;

    TASCoords cellToDraw = m_coordsSelCell.IsValid() ? m_coordsSelCell : m_coordsCurrCell;
    if (cellToDraw.IsValid()) {
      if (m_bShowFrontier && (g_pTileSpace->IsFrontierLocation(cellToDraw))) {
        if (g_pTileSpace->IsDeadFrontierLocation(cellToDraw)) {
          DrawTileEx(&dc, cellToDraw, NULL, TAS_DRAW_FRONTIERDEAD, true);
        } else {
          DrawTileEx(&dc, cellToDraw, NULL, TAS_DRAW_FRONTIER, true);
        }
      } else {
        DrawTileEx(&dc, cellToDraw, g_pTileSpace->GetTileType(cellToDraw), TAS_DRAW_TILE, true);
      }
    }

    if (m_parentFrame->m_pCellValueWin) {
      m_parentFrame->m_pCellValueWin->SetCurrentCell(cellToDraw, false, false);
    }
  }
  event.Skip();

  m_lBeginX = event.GetX();
  m_lBeginY = event.GetY();

  m_parentFrame->UpdateStatusBar();

  return;
}

void TASTilePanel::SetMouseMode(int mode)
{
  m_mouseMode = mode;
}

void TASTilePanel::ResetRubberband()
{
  m_RubberBand.ClearOldRubberRect();
  m_rubberbandorigin = wxPoint(INVALID_LOC, INVALID_LOC);
}

void TASTilePanel::BuildRtClickMenu(wxMenu* pMenu)
{
  for (size_t i=0; i<g_pTileSpace->m_tileset.size(); i++) {
    pMenu->Append(Menu_RightClick_TileAdd, g_pTileSpace->m_tileset[i]->m_sName);
  }
}

void TASTilePanel::OnRtClickMenu(wxCommandEvent& event)
{
/*
  wxString sFoo = event.GetString();
  if (sFoo.Len() > 0) {
    int blah = 1;
  }
  int dsfsa = event.GetInt();
*/
}

wxString TASTilePanel::GetPositionStatusText()
{
  if (!m_coordsCurrCell.IsValid()) {
    return wxString(_T(""));
  } else {
    return wxString::Format(_T("X:%d   Y:%d"),
                            m_coordsCurrCell.x,
                            m_coordsCurrCell.y);
  }
}

void TASTilePanel::UpdateSpace(bool bEntire)
{
  if (bEntire) {
///////////////////////////////////////////////////////////////////////
//MJP: get the DC for the canvas and call SetClippingRegion to be the entire viewable space
/*
    int xViewStart = 0;
    int yViewStart = 0;
    GetViewStart(&xViewStart, &yViewStart);

    int xScrollPixelsPerUnit = 0;
    int yScrollPixelsPerUnit = 0;
    GetScrollPixelsPerUnit(&xScrollPixelsPerUnit, &yScrollPixelsPerUnit);

    int xClientWidth  = 0;
    int yClientHeight = 0;
    GetClientSize(&xClientWidth, &yClientHeight);

    int virtWidth  = 0;
    int virtHeight = 0;
    GetVirtualSize(&virtWidth, &virtHeight);
*/
/*
    wxClientDC dc(this);
    DoPrepareDC(dc);

    dc.DestroyClippingRegion();

    //dc.SetClippingRegion((xViewStart * xScrollPixelsPerUnit), (yViewStart * yScrollPixelsPerUnit),
    //                     xClientWidth, yClientHeight);
    dc.SetClippingRegion(0, 0, virtWidth, virtHeight);
*/
//    RefreshRect(wxRect(0, 0, virtWidth, virtHeight));
//RefreshRect(wxRect(0, 0, virtWidth, virtHeight));

///////////////////////////////////////////////////////////////////////
    Refresh();
    m_parentFrame->RefreshOverviewWin();
  } else {
    DrawIteration(g_pTileSpace->m_epoch);
    if (m_parentFrame->m_pOverviewWin) {
      ((TASTileOverviewWin*)m_parentFrame->m_pOverviewWin)->DrawIteration(g_pTileSpace->m_epoch);
    }
  }
  m_parentFrame->UpdateStatusBar();
}

void TASTilePanel::OnKeyboard(wxKeyEvent& event)
{
  bool  bDoRot       = false;
  bool  bDoRefresh   = false;
  float phi          = RADS_IN_CIRCLE / 36.0f;
  float axis[3]      = {0.0f, 0.0f, 0.0f };
  float spin_quat[4] = {0.0f, 0.0f, 0.0f, 0.0f };
  
  if (event.m_callbackUserData) {
    wxPoint* pPt = (wxPoint*)event.m_callbackUserData;
    phi = phi * pPt->x;
  }

  if (event.GetKeyCode() == 'R') {    //  Do reset back to original orientation
//MJP: use the fact that this performs a 'reset' to figure out how to rotate around the actual axes above!
    bDoRefresh = true;
  } else if (event.GetKeyCode() == '+') {    //  Zoom in
    SetScaleFact(m_fScaleFact + 25.0f);
    bDoRefresh = true;
  } else if (event.GetKeyCode() == '-') {    //  Zoom out
    SetScaleFact(m_fScaleFact - 25.0f);
    bDoRefresh = true;
  }

  if (bDoRefresh) {
    //  Orientation has changed, redraw mesh
    Refresh();
  }
}

void TASTilePanel::OnScrollEx(wxScrollEvent& event)
{
  event.Skip();
}

void TASTilePanel::OnScrollWin(wxScrollWinEvent& event)
{

//  WXWORD vk = VK_MBUTTON;
//  SHORT foo = GetAsyncKeyState(vk);
//  wxMouseState ms = ::wxGetMouseState();
//  bool b = ms.MiddleDown();

//  This redraw is done automatically by TASTilePanel::ReDrawSpace
/*
  if (!m_bIsOverview && m_parentFrame->m_pOverviewWin) {
    m_parentFrame->m_pOverviewWin->Refresh();
  }
*/
  event.Skip();
}

void TASTilePanel::OnSize(wxSizeEvent& event)
{
  if (m_parentFrame->m_pOverviewWin) {
    ((TASTileOverviewWin*)m_parentFrame->m_pOverviewWin)->ClearOverlay(NULL);
  }
  event.Skip();
  if (m_parentFrame->m_pOverviewWin) {
    ((TASTileOverviewWin*)m_parentFrame->m_pOverviewWin)->DrawOverlay(NULL, false);
  }
}
