/////////////////////////////////////////////////////////////////////////////
// Name:        TASTilePanel.h
// Purpose:     Declares the canvas on which the 2-D tile space is drawn
// Author:      Matthew J. Patitz
// Copyright:   Copyright 2008, 2009 Matthew J. Patitz
// Licence:
//    This file is part of ISU TAS.
//
//    ISU TAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    ISU TAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with ISU TAS.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////////////

#ifndef _TAS_TILE_PANEL_H
#define _TAS_TILE_PANEL_H

class TASEditorFrame;

#include "wxprec.h"
#include "TASTiles.h"
#include "TASEditor.h"
#include "TASPanel.h"
#include <wx/clrpicker.h>


//  Forward declaration
class TASTileTypeWin;

//MJP: move TASIntVect out into TAS.h or something and also use for breakpoints and analyzerEpochs
class TASIntVect : public std::vector<int>
{
public:
  bool Contains(int i);
};
typedef TASIntVect::iterator TASIntVectIter;

class TASTileEditor : public wxScrolledWindow
{
public:
  TASTileEditor(wxWindow *parent,
                  TASEditorFrame *editorFrame,
                  wxWindowID id = -1,
                  const wxPoint& pos = wxDefaultPosition,
                  const wxSize& size = wxDefaultSize,
                  long style = wxWANTS_CHARS | wxScrolledWindowStyle,
                  const wxString& name = _T("Tile Editor"));

  void OnPaint(wxPaintEvent& event);
  void OnSize(wxSizeEvent& event);
  void OnMouse(wxMouseEvent& event);

  void OnEdit(wxCommandEvent& event);
  void OnInsert(wxCommandEvent& event);
  void OnCut(wxCommandEvent& event);
  void OnCopy(wxCommandEvent& event);
  void OnPaste(wxCommandEvent& event);
  void OnDelete(wxCommandEvent& event);
  void OnRotate(wxCommandEvent& event);
  void OnFind(wxCommandEvent& event);
  void OnFindBinding(wxCommandEvent& event);
  void OnSelectUsed(wxCommandEvent& event);
  void OnKey(wxKeyEvent& event);

  static void BuildEditMenu(wxMenu* pMenu);

  void SelectTile(int tileNum, wxDC* pDC);
  void UnSelectTile(int tileNum, wxDC* pDC);
  bool IsTileSelected(int tileNum);
  bool IsOverSelectedTile(wxPoint& pt);

  bool PutTileTypesOnClipboard(TASIntVect& ttNumList, bool bDelete = false);

  void GetTileLocation(int tileNum, int* x, int* y);    //  Returns screen coords of tiletype number (beginning point of drawn rect)
  int  GetTileFromLocation(int x, int y);               //  Returns tiletype num given screen coords
  int  GetTileFromGridLoc(int row, int col);            //  Returns tiletype num given row and column numbers

  void DrawTileEx(wxDC* pDC, int tileNum, int xOffset = 0, int yOffset = 0);

  wxString GenerateTileName(const wxChar* cNameAttempt);
  void FindTileType(wxString sName, bool bScrollToCenter = true);
  void ScrollToCenter(int tileNum);

  TASEditorFrame*     m_pEditorFrame;

  wxPoint             m_ptMousePos;
  wxPoint             m_ptDragStart;
  wxPoint             m_ptPreDragStart;

private:

  int                 m_tileSpacing;
  int                 m_tilesPerRow;
  int                 m_currTileTypeNum;
  int                 m_savedTileTypeNum;
  bool                m_bInDrag;
  bool                m_bInSelectDrag;
  wxString            m_sClipboardData;
  wxSize              m_size3DTiles;
  TASIntVect          m_selectedList;
  TASIntVect          m_savedList;
  
  DECLARE_CLASS(TASTileEditor)
  DECLARE_EVENT_TABLE()
};

const int ID_TAS_TTYPE_ED_WIN         = 800;
const int ID_TAS_TTIMAGE_BUTTON       = 801;
//  Note that ID's 810-870 are reserved for Edit Menu ID's as TASEditorFrame forwards those ID's to the active TASTileEditor
const int ID_TAS_TILEED_EDIT          = 810;
const int ID_TAS_TILEED_CUT           = 811;
const int ID_TAS_TILEED_COPY          = 812;
const int ID_TAS_TILEED_PASTE         = 813;
const int ID_TAS_TILEED_ROTATE90      = 814;
const int ID_TAS_TILEED_ROTATE180     = 815;
const int ID_TAS_TILEED_ROTATE270     = 816;
const int ID_TAS_TILEED_INSERT        = 817;
const int ID_TAS_TILEED_FINDBINDING_N = 818;
const int ID_TAS_TILEED_FINDBINDING_E = 819;
const int ID_TAS_TILEED_FINDBINDING_S = 820;
const int ID_TAS_TILEED_FINDBINDING_W = 821;
const int ID_TAS_TILEED_FIND          = 822;
const int ID_TAS_TILEED_SELECTUSED    = 823;
const int ID_TAS_TILEED_SELECTUNUSED  = 824;
const int ID_TAS_TILEED_DELETE        = 825;


#endif /* _TAS_TILE_PANEL_H */



